print("import: 'google.auth'")
import google.auth

print("import: 'google.auth.credentials'")
import google.auth.credentials

print("import: 'google.auth.transport.requests'")
import google.auth.transport.requests

print("import: 'google.longrunning'")
import google.longrunning

print("import: 'google.longrunning.operations_pb2'")
import google.longrunning.operations_pb2

print("import: 'google.protobuf'")
import google.protobuf

print("import: 'google.rpc'")
import google.rpc

print("import: 'google.api_core'")
import google.api_core

print("import: 'google.api_core.exceptions'")
import google.api_core.exceptions

print("import: 'google.api_core.future'")
import google.api_core.future

print("import: 'google.api_core.future.base'")
import google.api_core.future.base

print("import: 'google.api_core.future.polling'")
import google.api_core.future.polling

print("import: 'google.api_core.general_helpers'")
import google.api_core.general_helpers

print("import: 'google.api_core.page_iterator'")
import google.api_core.page_iterator

print("import: 'google.api_core.protobuf_helpers'")
import google.api_core.protobuf_helpers

print("import: 'google.api_core.retry'")
import google.api_core.retry

print("import: 'google.api_core.timeout'")
import google.api_core.timeout

